@@ -17,45 +17,38 @@ use constant DEBUG => $ENV{PERL_DISTRIBUTION_METADATA_DEBUG};
 my $SEP = qr{/|\\}; # path separater
 my $ARCHNAME = $Config{archname};
 
-our $VERSION = "0.02";
+our $VERSION = "0.03";
 
-my $CACHE_CORE_DISTRIBUTION = 1; # default cache on
-my %CACHE;
-sub cache_core_distribution {
-    my $class = shift;
-    if (@_) {
-        $CACHE_CORE_DISTRIBUTION = $_[0];
-        undef %CACHE unless $_[0];
-    } else {
-        $CACHE_CORE_DISTRIBUTION = 1;
-    }
+our $CACHE;
+
+sub new_from_file {
+    my ($class, $file, %option) = @_;
+    $class->_new(%option, _module => {file => $file});
 }
 
 sub new_from_module {
     my ($class, $module, %option) = @_;
-    my $inc = $option{inc} || \@INC;
-    $inc = $class->_abs_path($inc);
-    if ($option{fill_archlib}) {
-        $inc = $class->_fill_archlib($inc);
-    }
-    my $metadata = Module::Metadata->new_from_module($module, inc => $inc);
-    if ($metadata) {
-        $class->new_from_file($metadata->filename, inc => $inc);
-    } else {
-        bless {}, $class;
-    }
+    $class->_new(%option, _module => {name => $module});
 }
 
-sub new_from_file {
-    my ($class, $file, %option) = @_;
+sub _new {
+    my ($class, %option) = @_;
+    my $module = $option{_module};
     my $inc = $option{inc} || \@INC;
     $inc = $class->_abs_path($inc);
-    if ($option{fill_archlib}) {
-        $inc = $class->_fill_archlib($inc);
-    }
-    my $self = bless {}, $class;
+    $inc = $class->_fill_archlib($inc) if $option{fill_archlib};
+    my $metadata = $module->{file}
+        ? Module::Metadata->new_from_file($module->{file}, inc => $inc)
+        : Module::Metadata->new_from_module($module->{name}, inc => $inc);
 
-    my ($packlist, $files) = $class->_find_packlist($file, $inc);
+    my $self = bless {}, $class;
+    return $self unless $metadata;
+
+    $module->{file} = $metadata->filename;
+    $module->{name} = $metadata->name;
+    $module->{version} = $metadata->version;
+
+    my ($packlist, $files) = $class->_find_packlist($module->{file}, $inc);
     if ($packlist) {
         $self->{packlist} = $packlist;
         $self->{files}    = $files;
@@ -67,10 +60,6 @@ sub new_from_file {
     if ($main_module) {
         $self->{main_module} = $main_module;
         if ($main_module eq "perl") {
-            if ($CACHE_CORE_DISTRIBUTION && !$CACHE{core_packlist}) {
-                $CACHE{core_packlist} = $packlist;
-                $CACHE{core_files}    = $files;
-            }
             $self->{main_module_version} = $^V;
             $self->{main_module_file} = $^X;
             $self->{dist} = "perl";
@@ -85,16 +74,25 @@ sub new_from_file {
     }
 
     my $archlib = catdir($lib, $ARCHNAME);
-    my $metadata = Module::Metadata->new_from_module(
+    my $main_metadata = Module::Metadata->new_from_module(
         $main_module, inc => [$archlib, $lib]
     );
-    return $self unless $metadata;
 
-    $self->{main_module_version} = $metadata->version;
-    $self->{main_module_file} = $metadata->filename;
+    my ($find_module, $find_version);
+    if ($main_metadata) {
+        $self->{main_module_version} = $main_metadata->version;
+        $self->{main_module_file} = $main_metadata->filename;
+        $find_module = $main_metadata->name;
+        $find_version = $main_metadata->version;
+    } else {
+        $find_module = $module->{name};
+        $find_version = $module->{version};
+    }
 
-    my ($meta_directory, $install_json, $install_json_hash, $mymeta_json)
-        = $class->_find_meta($metadata->name, $metadata->version, catdir($archlib, ".meta"));
+    my ($meta_directory, $install_json, $install_json_hash, $mymeta_json) = $class->_find_meta(
+        $main_module, $find_module, $find_version,
+        catdir($archlib, ".meta")
+    );
     $self->{meta_directory}    = $meta_directory;
     $self->{install_json}      = $install_json;
     $self->{install_json_hash} = $install_json_hash;
@@ -150,35 +148,54 @@ sub _fill_archlib {
     \@out;
 }
 
-my $JSON = JSON->new;
+my $decode_install_json = sub {
+    my $file = shift;
+    my $content = do { open my $fh, "<", $file or next; local $/; <$fh> };
+    JSON::decode_json($content);
+};
+sub _decode_install_json {
+    my ($class, $file) = @_;
+    if ($CACHE) {
+        $CACHE->{install_json}{$file} ||= $decode_install_json->($file);
+    } else {
+        $decode_install_json->($file);
+    }
+}
 sub _find_meta {
-    my ($class, $module, $version, $dir) = @_;
+    my ($class, $main_module, $module, $version, $dir) = @_;
     return unless -d $dir;
 
-    # to speed up, first try distribution which just $module =~ s/::/-/gr;
-    my $naive = do { my $dist = $module; $dist =~ s/::/-/g; $dist };
+    my @install_json;
+    if ($CACHE and $CACHE->{install_json_collected}) {
+        @install_json = keys %{$CACHE->{install_json}};
+    } else {
+        @install_json = do {
+            opendir my $dh, $dir or die "opendir $dir: $!";
+            my @meta_dir = grep { !/^[.]{1,2}$/ } readdir $dh;
+            grep -f, map { catfile($dir, $_, "install.json") } @meta_dir;
+        };
+        if ($CACHE) {
+            $CACHE->{install_json}{$_} ||= undef for @install_json;
+            $CACHE->{install_json_collected}++;
+        }
+    }
 
-    # note: don't use glob(); it deffers depending on perl version
-    my @install_jsons = do {
-        opendir my $dh, $dir or die "opendir $dir: $!";
-        my @e = grep { !/^[.]{1,2}$/ } readdir $dh;
-        my @sort = (
-            (sort { $b cmp $a } grep {  /^$naive/ } @e),
-            (sort { $b cmp $a } grep { !/^$naive/ } @e),
-        );
-        grep -f, map { catfile($dir, $_, "install.json") } @sort;
-    };
+    # to speed up, first try distribution which just $module =~ s/::/-/gr;
+    my $naive = do { my $dist = $main_module; $dist =~ s/::/-/g; $dist };
+    @install_json = (
+        (sort { $b cmp $a } grep {  /^$naive/ } @install_json),
+        (sort { $b cmp $a } grep { !/^$naive/ } @install_json),
+    );
 
     my ($meta_directory, $install_json, $install_json_hash, $mymeta_json);
     INSTALL_JSON_LOOP:
-    for my $file (@install_jsons) {
-        my $content = do { open my $fh, "<:utf8", $file or next; local $/; <$fh> };
-        my $hash = $JSON->decode($content);
+    for my $file (@install_json) {
+        my $hash = $class->_decode_install_json($file);
 
         # name VS target ? When LWP, name is LWP, and target is LWP::UserAgent
         # So name is main_module!
         my $name = $hash->{name} || "";
-        next if $name ne $module;
+        next if $name ne $main_module;
         my $provides = $hash->{provides} || +{};
         for my $provide (sort keys %$provides) {
             if ($provide eq $module
@@ -212,51 +229,77 @@ sub _naive_packlist {
 # eg: OSX,
 # in .packlist: /var/folders/...
 # but /var/folders/.. is a symlink to /private/var/folders
+my $extract_files = sub {
+    my $packlist = shift;
+    [
+        map  { Cwd::abs_path($_) } grep { -f }
+        sort keys %{ ExtUtils::Packlist->new($packlist) || +{} }
+    ];
+};
 sub _extract_files {
     my ($class, $packlist) = @_;
-    map  { Cwd::abs_path($_) }
-    grep { -f }
-    sort keys %{ ExtUtils::Packlist->new($packlist) || +{} };
+    if ($CACHE) {
+        $CACHE->{packlist}{$packlist} ||= $extract_files->($packlist);
+    } else {
+        $extract_files->($packlist);
+    }
+}
+
+sub _core_packlist {
+    my ($self, $inc) = @_;
+    for my $dir (grep -d, @$inc) {
+        opendir my $dh, $dir or die "Cannot open dir $dir: $!\n";
+        my ($packlist) = map { catfile($dir, $_) } grep {$_ eq ".packlist"} readdir $dh;
+        return $packlist if $packlist;
+    }
+    return;
 }
 
 sub _find_packlist {
     my ($class, $module_file, $inc) = @_;
 
-    if ($CACHE_CORE_DISTRIBUTION && $CACHE{core_files}) {
-        if ( grep { $_ eq $module_file } @{ $CACHE{core_files} } ) {
-            DEBUG and warn "-> hit cache core packlist: $module_file\n";
-            return ($CACHE{core_packlist}, $CACHE{core_files});
+    if ($CACHE and my $core_packlist = $CACHE->{core_packlist}) {
+        my $files = $class->_extract_files($core_packlist);
+        if (grep {$module_file eq $_} @$files) {
+            return ($core_packlist, $files);
         }
     }
 
     # to speed up, first try packlist which is naively guessed by $module_file
     if (my $naive_packlist = $class->_naive_packlist($module_file, $inc)) {
-        my @files = $class->_extract_files($naive_packlist);
-        if ( grep { $module_file eq $_ } @files ) {
+        my $files = $class->_extract_files($naive_packlist);
+        if ( grep { $module_file eq $_ } @$files ) {
             DEBUG and warn "-> naively found packlist: $module_file\n";
-            return ($naive_packlist, \@files);
+            return ($naive_packlist, $files);
         }
     }
 
-    my ($packlist, $files);
-    for my $dir ( grep -d, map {(catdir($_, "auto"), $_)} @{ $class->_fill_archlib($inc) } ) {
-        last if $packlist;
-        find {
-            wanted => sub {
-                return if $packlist;
-                return unless -f $_ && basename($_) eq ".packlist";
-                return if $CACHE_CORE_DISTRIBUTION && ($CACHE{core_packlist} || "") eq $_;
-                my @files = $class->_extract_files($_);
-                if ( grep { $module_file eq $_ } @files ) {
-                    $packlist = $File::Find::name;
-                    $files = \@files;
-                    return;
-                }
-            },
-            no_chdir => 1,
-        }, $dir;
+    my @packlists;
+    if ($CACHE and $CACHE->{packlist_collected}) {
+        @packlists = keys %{ $CACHE->{packlist} };
+    } else {
+        if (my $core_packlist = $class->_core_packlist($inc)) {
+            push @packlists, $core_packlist;
+            $CACHE->{core_packlist} = $core_packlist if $CACHE;
+        }
+        find sub {
+            return unless -f;
+            return unless $_ eq ".packlist";
+            push @packlists, $File::Find::name;
+        }, grep -d, map { catdir($_, "auto") } @{$class->_fill_archlib($inc)};
+        if ($CACHE) {
+            $CACHE->{packlist}{$_} ||= undef for @packlists;
+            $CACHE->{packlist_collected}++;
+        }
     }
-    return ($packlist, $files);
+
+    for my $try (@packlists) {
+        my $files = $class->_extract_files($try);
+        if (grep { $module_file eq $_ } @$files) {
+            return ($try, $files);
+        }
+    }
+    return;
 }
 
 sub _abs_path {
@@ -528,7 +571,7 @@ L<App::cpanminus>
 
 =head1 LICENSE
 
-Copyright (C) Shoichi Kaji.
+Copyright (C) 2015 Shoichi Kaji
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
